print("import: 'lunr'")
import lunr

print("import: 'lunr.builder'")
import lunr.builder

print("import: 'lunr.exceptions'")
import lunr.exceptions

print("import: 'lunr.field_ref'")
import lunr.field_ref

print("import: 'lunr.idf'")
import lunr.idf

print("import: 'lunr.index'")
import lunr.index

print("import: 'lunr.languages'")
import lunr.languages

print("import: 'lunr.languages.stemmer'")
import lunr.languages.stemmer

print("import: 'lunr.languages.trimmer'")
import lunr.languages.trimmer

print("import: 'lunr.match_data'")
import lunr.match_data

print("import: 'lunr.pipeline'")
import lunr.pipeline

print("import: 'lunr.query'")
import lunr.query

print("import: 'lunr.query_lexer'")
import lunr.query_lexer

print("import: 'lunr.query_parser'")
import lunr.query_parser

print("import: 'lunr.stemmer'")
import lunr.stemmer

print("import: 'lunr.stop_word_filter'")
import lunr.stop_word_filter

print("import: 'lunr.token'")
import lunr.token

print("import: 'lunr.token_set'")
import lunr.token_set

print("import: 'lunr.token_set_builder'")
import lunr.token_set_builder

print("import: 'lunr.tokenizer'")
import lunr.tokenizer

print("import: 'lunr.trimmer'")
import lunr.trimmer

print("import: 'lunr.utils'")
import lunr.utils

print("import: 'lunr.vector'")
import lunr.vector

